# Copyright The IETF Trust 2017, All Rights Reserved
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function

series_name = {
    'std':  'STD',
    'bcp':  'BCP',
    'info': 'FYI',
}

category_name = {
    'rfc':      'Request for Comments',
    'std':      'Standards Track',
    'bcp':      'Best Current Practice',
    'exp':      'Experimental',
    'info':     'Informational',
    'historic': 'Historic',
}

stream_name = {
    'IETF':         'Internet Engineering Task Force (IETF)',
    'IAB':          'Internet Architecture Board (IAB)',
    'IRTF':         'Internet Research Task Force (IRTF)',
    'independent':  'Independent Submission',
    'editorial':    'Editorial Stream',
}



# Copyright The IETF Trust 2017, All Rights Reserved
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function

import i18naddress
import lxml.etree
import pycountry
import re
import xml2rfc.log

try:
    from xml2rfc import debug
    debug.debug = True
except ImportError:
    pass

from xml2rfc.util.name import full_author_name, full_author_ascii_name, full_org_name, full_org_ascii_name

address_field_mapping = dict(i18naddress.FIELD_MAPPING)
address_field_mapping.update({'E':'extended_address', 'Y':'country_name'})

address_field_elements = {
    # mapping from address fields to element tags
#    'name':             'fullname',
#    'company_name':     'organization',
    'extended_address': '<extaddr>',
    'street_address':   '<street>',
    'sorting_code':     '<sortingcode>',
    'postal_code':      '<code>',
    'city_area':        '<cityarea>',
    'city':             '<city>',
    'country_area':     '<region>',
    'country_name':     '<country>',
}

address_hcard_properties = {
    'name':             'fn nameRole',
    'company_name':     'org',
    'extended_address': 'extended-address',
    'street_address':   'street-address',
    'sorting_code':     'postal-code',
    'postal_code':      'postal-code',
    'city_area':        'locality',
    'city':             'locality',
    'country_area':     'region',
    'country_name':     'country-name'
}

def get_value(e, latin=True):
    if latin:
        value = e.get('ascii') or e.text or ''
    else:
        value = e.text or ''
    return value

country_alias = {
    "UK": "United Kingdom",
    "Republic of Korea": "Korea, Republic of",
    "South Korea": "Korea, Republic of",
    "North Korea": "Democratic People's Republic of Korea",
    "P.R. China": "China",
    "Åland": "Åland Islands",
    # Reordering of comma-separated name components:
    "The Democratic Republic of the Congo": "Congo, The Democratic Republic of the",
    "State of Palestine": "Palestine, State of",
    "U.S. Virgin Islands": "Virgin Islands, U.S.",
}

def get_iso_country_info(e):
    country_info = None
    ascii = e.get('ascii')
    if ascii:
        name = ascii.strip()
    elif e.text:
        name = e.text.strip()
    else:
        return None
    # Special case, this is used but unknown to pycountry
    if name in country_alias:
        name = country_alias[name]
    try:
        if hasattr(pycountry.countries, 'lookup'):
            country_info = pycountry.countries.lookup(name)
        else:
            # Workaround for old versions of pycountry.  The transition from
            # using alpha2 to alpha_2 coincides with the presence of a
            # lookup() method for regular releases (there's a discrepancy for
            # rc versions 16.10.23rc*)
            if len(name) == 2:
                country_info = pycountry.countries.get(alpha2=name.upper())
            else:
                country_info = pycountry.countries.get(name=name.title())
            country_info.alpha_2 = country_info.alpha2
    except LookupError:
        estr = lxml.etree.tostring(e).replace(b' xmlns:xi="http://www.w3.org/2001/XInclude"', b'').strip().decode('utf-8')
        if e.tag == 'country':
            xml2rfc.log.warn("Country lookup failed for %s.  Use --country-help to show recognized country names and codes" % (estr, ))
        else:
            xml2rfc.log.note("Country lookup failed for %s.  Use --country-help to show recognized country names and codes" % (estr, ))
        pass
    return country_info

def get_normalized_address_info(writer, x, latin=False):
    author = x.getparent().getparent()
    name = full_author_ascii_name(author) if latin else full_author_name(author)
    role = author.get('role')
    company = full_org_ascii_name(author) if latin else full_org_name(author)
    country_info = None
    country_element = x.find('country')
    if country_element != None and country_element.text:
        country_info = get_iso_country_info(country_element)
    if not country_info:
        for c in x.getchildren():
            if c.tag == "postalLine":
                country_info = get_iso_country_info(c)
                if country_info:
                    country_element = c
                    break
    if country_info:
        country_name = get_value(country_element, latin=latin)
        if country_name in ['US', 'USA']:
            country_name = country_info.official_name
        elif re.match('[A-Z]{2,3}', country_name):
            country_name = country_info.name
        adr = {
                'name': name,
                'role': role,
                'company_name': company,
                'extended_address': [],
                'street_address': [],
                'sorting_code': '',
                'postal_code': '',
                'city_area': '',
                'city': '',
                'country_area': '',
                'country_code': country_info.alpha_2,
                'country_name': country_name,
            }
    else:
        adr = {
                'name': name,
                'role': role,
                'company_name': company,
                'extended_address': [],
                'street_address': [],
                'sorting_code': '',
                'postal_code': '',
                'city_area': '',
                'city': '',
                'country_area': '',
                'country_code': '',
                'country_name': ''
            }
    for c in x.getchildren():
        if c == country_element:
            continue
        # Some of these will overwrite data if there are multiple elements
        value = get_value(c, latin=latin)
        if value:
            if   c.tag == 'extaddr':
                adr['extended_address'].append(value)
            elif c.tag in ['street', 'postalLine', 'pobox']:
                adr['street_address'].append(value)
            elif c.tag == 'cityarea':
                adr['city_area'] = value
            elif c.tag == 'city':
                adr['city'] = value
            elif c.tag == 'region':
                adr['country_area'] = value
            elif c.tag == 'code':
                adr['postal_code'] = value
            elif c.tag == 'sortingcode':
                adr['sorting_code'] = value
    for a in adr:
        if isinstance(adr[a], list):
            adr[a] = '\u2028'.join(adr[a])
    if country_info:
        # Address validation
        address_format, rules = get_address_format_rules(adr, latin)
        parts = [ address_field_mapping[c] for c in re.findall(r'%([A-Z])', address_format)  if not c in ['N', 'O'] ]
        elements = [ address_field_elements[p] for p in parts ]
        list_parts = False
        for e in adr:
            if adr[e] and not (e in parts or e in ['name', 'role', 'company_name', 'country_code', ]):
                list_parts = True
                writer.warn(x, "Postal address part filled in, but not used: %s: %s\n"
                               " Available elements for %s are: %s" % (address_field_elements[e], adr[e], country_info.name, ', '.join(elements)))
        try:
            i18naddress.normalize_address(adr)
        except i18naddress.InvalidAddressError as e:
            list_parts = True
            writer.note(x, "Postal address check failed for author %s." % full_author_name(author, latin))
            for item in e.errors:
                if adr[item]:
                    writer.note(x, "  Postal address has an unexpected address field or field value: %s: %s" % (address_field_elements[item], (adr[item])), label='')
                else:
                    writer.note(x, "  Postal address is missing an address element: %s" % (address_field_elements[item], ), label='')
        if list_parts:
            writer.note(x, "Recognized postal address elements for %s are: %s" % (rules.country_name.title(), ', '.join(elements)))
    return adr

# These are copied from i18address in order to remove uppercasing


def get_address_format_rules(address, latin=False, normalize=False):
    rules = i18naddress.get_validation_rules(address)
    address_format = rules.address_latin_format if latin else rules.address_format
    if rules.address_latin_format != rules.address_format:
        # don't normalize the line order if the country provides
        # a latin-specific address format
        normalize = False
    address_format = enhance_address_format(address, address_format)
    if normalize:
        address_format = normalize_address_format(address_format)
    return address_format, rules

def enhance_address_format(adr, address_format):
    #     'A': 'street_address'
    #     'C': 'city'
    #     'D': 'city_area'
    #     'E': 'extended_address'
    #     'N': 'name'
    #     'O': 'company_name'
    #     'S': 'country_area'
    #     'X': 'sorting_code'
    #     'Y': 'country_name '
    #     'Z': 'postal_code'

    # Add extended address field '%E' (the code fails for 4 countries: Guinea,
    # Hungary, Iran and China, which won't have %E support):
    address_format = address_format.replace('%N%n%O%n%A', '%N%n%O%n%E%n%A')
    address_format = address_format.replace('%A%n%O%n%N', '%A%n%E%n%O%n%N')
    address_format = address_format.replace('%O%n%N%n%A', '%O%n%N%n%E%n%A')
    if '%E%n%O%n%N' in address_format or '%A%n%O%n%N' in address_format:
        address_format = '%Y%n' + address_format
    else:
        address_format = address_format + '%n%Y'
    # This is incorrect, but suits us better when listing authors
    # in the Authors' Addresses section: move the name to the first line
    address_format = '%N%n' + re.sub(r'%N(%n)?', '', address_format)
    # Get rid of any trailing newlines we might have created
    address_format = re.sub(r'%n$', '', address_format)
    # Add some spacing, for a few special cases
    address_format = re.sub(r'%([A-Z])%([A-Z])', r'%\1 %\2', address_format)
    address_format = re.sub(r'%([A-Z])%([A-Z])', r'%\1 %\2', address_format)
    # country-specific fixes, if any
    if adr['country_code'] == 'SE':
        if not adr['postal_code']:
            address_format = re.sub(r'SE-%Z ', r'', address_format)            
    return address_format

    
def normalize_address_format(address_format):
    #     'N': 'name'
    #     'O': 'company_name'
    #     'E': 'extended_address'
    #     'A': 'street_address'
    #     'X': 'sorting_code'
    #     'D': 'city_area'
    #     'C': 'city'
    #     'Z': 'postal_code'
    #     'S': 'country_area'
    #     'Y': 'country_name '
    old = address_format
    normalized_order = list("NOEAXDCZSY")
    lines = address_format.split('%n')
    norm = []
    # iterate through the desired field order, but don't mess with the address
    # element order within a line:
    for code in normalized_order:
        code = '%'+code
        found = None
        for i, line in enumerate(lines):
            if code in line:
                norm.append(line)
                found = i
                break
        if found:
            del lines[found]
    address_format = '%n'.join(norm)
    new = address_format
    if old != new:
        debug.show('old')
        debug.show('new')
    return address_format



# Copyright The IETF Trust 2018, All Rights Reserved
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, division

import io
import logging
import os
import re


try:
    import weasyprint
    import_error = None
except (ImportError, OSError, ValueError) as e:
    import_error = e
    weasyprint = False


import xml2rfc
from xml2rfc.writers.base import default_options, BaseV3Writer
from xml2rfc.writers.html import HtmlWriter
from xml2rfc.util.fonts import get_noto_serif_family_for_script


try:
    from xml2rfc import debug
    debug.debug = True
except ImportError:
    pass


NOTO_SYMBOLS = ["NotoSansSymbols2", "NotoSansMath",]


class PdfWriter(BaseV3Writer):

    def __init__(self, xmlrfc, quiet=None, options=default_options, date=None):
        super(PdfWriter, self).__init__(xmlrfc, quiet=quiet, options=options, date=date)
        if not weasyprint:
            self.err(None, "Cannot run PDF formatter: %s" % import_error)
            return

        logging.basicConfig(level=logging.INFO)

        loggers = []
        loggers.append(logging.getLogger("weasyprint"))
        loggers.append(logging.getLogger("fontTools.subset"))
        loggers.append(logging.getLogger("fontTools.ttLib"))
        loggers.append(logging.getLogger("fontTools.varlib"))
        for _logger in loggers:
            if   self.options.quiet:
                _logger.setLevel(logging.CRITICAL)
            elif self.options.verbose:
                _logger.setLevel(logging.WARNING)
            elif self.options.debug:
                _logger.setLevel(logging.DEBUG)
            else:
                _logger.setLevel(logging.ERROR)

    def pdf(self):
        if not weasyprint:
            return None

        if not self.root.get('prepTime'):
            prep = xml2rfc.PrepToolWriter(self.xmlrfc, options=self.options, date=self.options.date, liberal=True, keep_pis=[xml2rfc.V3_PI_TARGET])
            tree = prep.prep()
            self.tree = tree
            self.root = self.tree.getroot()

        self.options.no_css = True
        self.options.pdf = True
        htmlwriter = HtmlWriter(self.xmlrfc, quiet=True, options=self.options, date=self.date)
        html = htmlwriter.html()

        html = self.flatten_unicode_spans(html)

        writer = weasyprint.HTML(string=html, base_url="")

        cssin  = self.options.css or os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'xml2rfc.css')
        css = weasyprint.CSS(cssin)

        # fonts and page info
        fonts = self.get_serif_fonts()
        mono_fonts = self.get_mono_fonts()
        page_info = {
            'top-left': self.page_top_left(),
            'top-center': self.full_page_top_center(),
            'top-right': self.page_top_right(),
            'bottom-left': self.page_bottom_left(),
            'bottom-center': self.page_bottom_center(),
            'fonts': ', '.join(fonts),
            'mono-fonts': ', '.join(mono_fonts),
        }
        for (k,v) in page_info.items():
            page_info[k] = v.replace("'", r"\'")
        page_css_text = page_css_template.format(**page_info)
        page_css = weasyprint.CSS(string=page_css_text)

        pdf = writer.write_pdf(None, stylesheets=[ css, page_css ], presentational_hints=True)

        return pdf


    def write(self, filename):
        if not weasyprint:
            return

        self.filename = filename

        pdf = self.pdf()
        if pdf:
            with io.open(filename, 'bw') as file:
                file.write(pdf)
            
            if not self.options.quiet:
                self.log(' Created file %s' % filename)
        else:
            self.err(None, 'PDF creation failed')

    def get_serif_fonts(self):
        fonts = set()
        scripts = self.root.get("scripts").split(",")
        noto_serif = "Noto Serif"
        for script in scripts:
            family = get_noto_serif_family_for_script(script)
            fonts.add("%s" % family)
        fonts -= set([ noto_serif, ])
        fonts = [noto_serif,] + NOTO_SYMBOLS + list(fonts)
        self.note(None, "Found installed font: %s" % ", ".join(fonts))
        return fonts

    def get_mono_fonts(self):
        fonts = set()
        scripts = self.root.get('scripts').split(',')
        roboto_mono = "Roboto Mono"
        noto_sans_mono = "Noto Sans Mono"
        for script in scripts:
            family = get_noto_serif_family_for_script(script)
            fonts.add("%s" % family)
        fonts = [ roboto_mono, noto_sans_mono, ] + NOTO_SYMBOLS + list(fonts)
        self.note(None, "Found installed font: %s" % ', '.join(fonts))
        return fonts

    def flatten_unicode_spans(self, html):
        # This is a fix for bug in WeasyPrint that doesn't handle RTL unicode
        # content correctly.
        # See #873 & Kozea/WeasyPrint#1711
        return re.sub(r'<span class="unicode">(?P<unicode_content>.*?)</span>',
                      r'\g<unicode_content>',
                      html)


page_css_template = """
@media print {{
  body {{
    font-family: {fonts};
    width: 100%;
  }}
  tt, code, pre {{
    font-family: {mono-fonts};
  }}
  sup {{
    line-height: 0;
  }}
  @page {{
    size: A4;
    font-size: 12px; /* needed for the page header and footer text */
    font-family: {fonts};

    border-top: solid 1px #ccc;
    padding-top: 18px;

    padding-bottom: 24px;
    border-bottom: solid 1px #ccc;
    margin-bottom: 45mm;

    @top-left {{
      content: '{top-left}';
      vertical-align: bottom;
      border-bottom: 0px;
      margin-bottom: 1em;
    }}
    @top-center {{
      content: '{top-center}';
      vertical-align: bottom;
      border-bottom: 0px;
      margin-bottom: 1em;
    }}
    @top-right {{
      content: '{top-right}';
      vertical-align: bottom;
      border-bottom: 0px;
      margin-bottom: 1em;
    }}
    @bottom-left {{
      content: '{bottom-left}';
      vertical-align: top;
      border-top: 0px;
      margin-top: 1em;
    }}
    @bottom-center {{
      content: '{bottom-center}';
      vertical-align: top;
      border-top: 0px;
      margin-top: 1em;
    }}
    @bottom-right {{
      content: 'Page ' counter(page);
      vertical-align: top;
      border-top: 0px;
      margin-top: 1em;
    }}
  }}
  #toc nav ul li p a:nth-child(2)::after {{
    content: target-counter(attr(href), page);
    position: absolute;
    right: 0px;
    text-indent: 0px;
  }}
}}
"""





# --------------------------------------------------
# Copyright The IETF Trust 2011, All Rights Reserved
# --------------------------------------------------

# Python libs
import textwrap
import lxml
import re

try:
    from xml2rfc import debug
    assert debug
except ImportError:
    pass

# Local lib
from xml2rfc.writers.base import BaseRfcWriter, default_options
from xml2rfc.util.unicode import expand_unicode_element
import xml2rfc.utils


class RawTextRfcWriter(BaseRfcWriter):
    """ Writes to a text file, unpaginated, no headers or footers.

        The page width is controlled by the *width* parameter.
    """

    def __init__(self, xmlrfc, width=72, margin=3, quiet=None, options=default_options, date=None):
        if not quiet is None:
            options.quiet = quiet
        BaseRfcWriter.__init__(self, xmlrfc, options=options, date=date)
        # Document processing data
        self.width = width      # Page width
        self.margin = margin    # Page margin
        self.buf = []           # Main buffer during processing
        self.output = []        # Final buffer that gets written to disk
        self.toc_marker = 0     # Line number in buffer to write toc to
        self.iref_marker = 0    # Line number in buffer to write index to
        self.list_counters = {} # Maintain counters for 'format' type lists
        self.edit_counter = 0   # Counter for edit marks
        # Set this to False to permit utf+8 output:
        self.ascii = not options.utf8       # Enable ascii flag
        self.cref_counter = 0   # Counter for cref anchors
        self.cref_list = []

        # Marks position of iref elements
        self.iref_marks = {}

        # Text lookups
        self.inline_tags = ['xref', 'eref', 'iref', 'cref', 'spanx', 'u', ]
        
        # Custom textwrapper object
        self.wrapper = xml2rfc.utils.TextWrapper(width=self.width,
                                                   fix_sentence_endings=True)

    def _length(self, text):
        return len(text)

    def _lb(self, buf=None, text='', num=1):
        """ Write a blank line to the file, with optional filler text 
        
            Filler text is usually used by editing marks
        """
        if num > 0:
            if buf is None:
                buf = self.buf
            buf.extend([text] * num)

    def _vspace(self, num=0):
        """ <vspace> line break wrapper to allow for overrides """
        return self._lb(num=num)

    def write_text(self, string, indent=0, sub_indent=0, bullet='',
                  align='left', leading_blankline=False, buf=None,
                  strip=True, edit=False, wrap_urls=True,
                  fix_sentence_endings=True, source_line=None,
                  fix_doublespace=True):
        """ Writes a line or multiple lines of text to the buffer.

            Several parameters are included here.  All of the API calls
            for text writers use this as the underlying method to write data
            to the buffer, with the exception of write_raw() that handles
            preserving of whitespace.
        """
        if buf is None:
            buf = self.buf

        if leading_blankline:
            if edit and self.pis['editing'] == 'yes':
                # Render an editing mark
                self.edit_counter += 1
                self._lb(buf=buf, text=str('<' + str(self.edit_counter) + '>'))
            else:
                self._lb(buf=buf)

        # We can take advantage of textwrap's initial_indent by using a bullet
        # parameter and treating it separately.  We still need to indent it.
        subsequent = ' ' * (indent + sub_indent)
        if bullet:
            initial = ' ' * indent
            bullet_parts = self.wrapper.wrap(bullet, initial=initial,
                                             fix_doublespace=False, drop_whitespace=False)
            if len(bullet_parts) > 1:
                buf.extend(bullet_parts[:-1])
            initial = bullet_parts[-1]
            if not sub_indent:
                # Use bullet length for subsequent indents
                subsequent = ' ' * len(initial)
        else:
            # No bullet, so combine indent and sub_indent
            initial = subsequent

        if string:
            if strip:
                # Strip initial whitespace
                string = string.lstrip()
            if wrap_urls:
                par = self.wrapper.wrap(string, initial=initial, subsequent_indent=subsequent,
                                        fix_sentence_endings=fix_sentence_endings,
                                        fix_doublespace=fix_doublespace)
            else:
                par = self.wrapper.wrap(xml2rfc.utils.urlkeep(string), initial=initial,
                                        subsequent_indent=subsequent,
                                        fix_sentence_endings=fix_sentence_endings,
                                        fix_doublespace=fix_doublespace)
            if len(par) == 0 and bullet:
                par = [ initial ]                    
            if align == 'left':
                buf.extend(par)
            elif align == 'center':
                for line in par:
                    margin = ' ' * indent
                    if line.startswith(margin):
                        line = line.rstrip()
                        centered = ' ' * ((self.width - len(line.rstrip()))//2)
                        centered += line
                        # centered = margin + line[len(margin):].center(self.width-indent-4).rstrip()
                        buf.append(centered)
                    else:
                        buf.append(line.center(self.width).rstrip())
            elif align == 'right':
                for line in par:
                    buf.append(line.rjust(self.width))
        elif bullet:
            # If the string is empty but a bullet was declared, just
            # print the bullet
            buf.append(initial)

    def write_list(self, list, level=0, indent=3):
        """ Writes a <list> element """
        bullet = '   '
        hangIndent = None
        # style comes from the node if one exists
        style = list.attrib.get('style', '')
        if not style:
            # otherwise look for the nearest list parent with a style and use it
            for parent in list.iterancestors():
                if parent.tag == 'list':
                    style = parent.attrib.get('style', '')
                    if style:
                        break
        if not style:
            style = 'empty'
        listlength = len(list.findall('t'))        
        if style == 'hanging' or style.startswith('format'):
            # Check for optional hangIndent
            try:
                hangIndent = list.attrib.get('hangIndent', None)
                if hangIndent != None:
                    hangIndent = int(hangIndent)
            except (ValueError, TypeError):
                xml2rfc.log.error("hangIndent value '%s' is not an integer" % hangIndent)
                hangIndent = 6
        format_str = None
        counter_index = None
        if style.startswith('format'):
            format_str = style.partition('format ')[2]
            allowed_formats = ('%c', '%C', '%d', '%i', '%I', '%o', '%x', '%X')
            if not any(map(lambda format: format in format_str, allowed_formats)):
                xml2rfc.log.warn('Invalid format specified: %s ' 
                                 '(Must be one of %s)' % (style,
                                    ', '.join(allowed_formats)))
            counter_index = list.attrib.get('counter', None)
            if not counter_index:
                counter_index = 'temp' + str(level)
                self.list_counters[counter_index] = 0
            elif counter_index not in self.list_counters:
                # Initialize if we need to
                self.list_counters[counter_index] = 0
        t_count = 0
        for element in list:
            # Check for PI
            if element.tag is lxml.etree.PI:
                pidict = self.parse_pi(element)
                if pidict and "needLines" in pidict:
                    self.needLines(pidict["needLines"])
            elif element.tag == 't':
                # Disable linebreak if subcompact=yes AND not first list element
                leading_blankline = True
                if t_count > 0 and element.pis['subcompact'] == 'yes':
                    leading_blankline = False
                if style == 'symbols':
                    bullet = element.pis['text-list-symbols'][level % len(element.pis['text-list-symbols'])]
                    bullet += '  '
                elif style == 'numbers':
                    bullet = self._format_counter("%d.", t_count+1, listlength)
                elif style == 'letters':
                    letter_style = "%C." if (level % 2) else "%c."
                    bullet = self._format_counter(letter_style, t_count+1, listlength)
                elif style == 'hanging':
                    bullet = element.attrib.get('hangText', '')
                    if hangIndent is None:
                        hangIndent = 3
                    if len(bullet) < hangIndent:
                        # Insert whitespace up to hangIndent
                        bullet = bullet.ljust(hangIndent)
                    else:
                        # Insert a single space
                        bullet += '  '
                    # Add an extra space in front of colon if colonspace enabled
                    if bullet.endswith(':') and \
                    element.pis['colonspace'] == 'yes':
                        bullet+= ' '
                    if element.text and len(bullet) > self.width//2:
                        # extra check of remaining space if the bullet is
                        # very long
                        first_word = self.wrapper._split(element.text)[0]
                        if len(first_word) > (self.width - len(bullet) - indent):
                            self.write_text('', bullet=bullet, indent=indent,
                                leading_blankline=leading_blankline)
                            self._vspace()
                            leading_blankline=False
                            indent = hangIndent
                            bullet = ''
                elif style.startswith('format'):
                    self.list_counters[counter_index] += 1
                    count = self.list_counters[counter_index]
                    bullet = self._format_counter(format_str, count, listlength)
                if hangIndent:
                    sub_indent = hangIndent
                else:
                    sub_indent = len(bullet)
                anchor = element.attrib.get('anchor')
                if anchor and self.indexmode:
                    self._indexListParagraph(t_count+1, anchor)
                self.write_t_rec(element, bullet=bullet, indent=indent, \
                                 level=level + 1, \
                                 sub_indent=sub_indent, leading_blankline=leading_blankline)
                t_count += 1

        
    def pre_write_toc(self):
        return ['', 'Table of Contents', '']

    def post_write_toc(self):
        return []

    def write_toc(self, paging=False):
        """ Write table of contents to a temporary buffer and return """
        if self.toc_marker < 1:
            # Toc is either disabled, or the pointer was messed up
            return ['']
        tmpbuf = []
        tmpbuf.extend(self.pre_write_toc())
        # Retrieve toc from the index
        tocindex = self._getTocIndex()
        tocdepth = self.pis['tocdepth']
        try:
            tocdepth = int(tocdepth)
        except ValueError:
            xml2rfc.log.warn('Invalid tocdepth specified, must be integer:', \
                             tocdepth)
            tocdepth = 3
        indent_scale = 2
        if self.pis['tocnarrow'] == 'no':
            indent_scale = 3
        for item in tocindex:
            # Add decoration to counter if it exists, otherwise leave empty
            if item.level <= tocdepth:
                counter = ''
                if item.counter:
                    counter = item.counter + '. '
                    # Extra space on single digit counters
                    if len(item.counter.rsplit('.')[-1]) == 1:
                        counter += ' '
                # Get item depth based on its section 'level' attribute
                depth = item.level - 1
                if depth < 0 or self.pis['tocindent'] == 'no':
                    depth = 0
                # Prepend appendix at first level
                if item.level == 1 and item.appendix:
                    counter = "Appendix " + counter
                bullet = ' ' * (depth * indent_scale) + counter
                indent = 3
                sub_indent = indent + len(bullet)
                pagestr = '%4s' % item.page
                lines = textwrap.wrap(bullet + (item.title.strip() if item.title else ""),
                                      self.width - len(pagestr),
                                      initial_indent=' ' * indent,
                                      subsequent_indent=' ' * sub_indent)
                if paging:
                    # Construct dots
                    dots = len(lines[-1]) % 2 and ' ' or ''
                    dots += ' .' * int((self.width - len(lines[-1]) - len(dots) + 1)//2)
                    lines[-1] += dots
                    # Insert page
                    lines[-1] = lines[-1][:0 - len(pagestr)] + pagestr
                tmpbuf.extend(lines)
        tmpbuf.extend(self.post_write_toc())
        return tmpbuf

    def pre_write_iref_index(self):
        return ['', 'Index']

    def write_iref_index(self):
        """ Write iref index to a temporary buffer and return """
        def pagelist(pages):
            pages = list(set(pages))
            pages.sort()
            # find ranges
            items = []
            prev = None
            for page in pages:
                if prev and (page - prev == 1):
                    items[-1].append(page)
                else:
                    items.append([page])
                prev = page
            pagelist = []
            for item in items:
                if len(item) == 1:
                    pagelist.append("%d"%item[0])
                else:
                    pagelist.append("%d-%d"%(item[0], item[-1]))
            return ", ".join(pagelist)

        if self.iref_marker < 1:
            # iref is either disabled, or the pointer was messed up
            return ['']
        tmpbuf = self.pre_write_iref_index()
        # Sort iref items alphabetically, store by first letter 
        alpha_bucket = {}
        keys = list(self._iref_index.keys())
        keys.sort(key=str.upper)
        for key in keys:
            letter = key[0].upper()
            if letter in alpha_bucket:
                alpha_bucket[letter].append(key)
            else:
                alpha_bucket[letter] = [key]
        for letter in sorted(alpha_bucket.keys()):
            # Write letter
            self.write_text(letter, indent=3, leading_blankline=True, buf=tmpbuf)
            for item in sorted(alpha_bucket[letter]):
                pages = self._iref_index[item].pages
                # Write item
                self.write_text(item + '  ' + pagelist(pages)
                                                        , indent=6, buf=tmpbuf, fix_doublespace=False)
                subkeys = list(self._iref_index[item].subitems.keys())
                subkeys.sort(key=str.upper)
                for subitem in subkeys:
                    pages = set(self._iref_index[item].subitems[subitem].pages)
                    # Write subitem
                    self.write_text(subitem + '  ' + pagelist(pages)
                                                        , indent=9, buf=tmpbuf, fix_doublespace=False)
        return tmpbuf

    def _expand_xref(self, xref):
        """ Returns the proper text representation of an xref element """
        target = xref.attrib.get('target', '')
        format = xref.attrib.get('format', self.defaults['xref_format'])
        item = self._getItemByAnchor(target)
        if not self.indexmode:
            if not item:
                xml2rfc.log.warn("Can't resolve xref target %s" % target)
            else:
                item.used = True
        if not item:
            target_text = '[' + target + ']'
        elif format == 'none':
            if xref.text:
                return xref.text.rstrip()
            return ''
        elif format == 'counter':
            target_text = item.counter
        elif format == 'title':
            target_text = item.title.strip()
        else: #Default
            target_text = item.autoName
        target_text = re.sub("-", u"\u2011", target_text) # switch to non-breaking hyphens
        target_text = re.sub(" ", u"\u00A0", target_text)  # switch to non-breaking space
        if xref.text:
            if not target_text.startswith('['):
                target_text = '(' + target_text + ')'
            return xref.text.rstrip() + ' ' + target_text
        else:
            return target_text

    def _expand_u(self, u):
        try:
            text = expand_unicode_element(u)
        except (RuntimeError, ValueError) as exc:
            text = ''
            xml2rfc.log.error('%s'%exc)
        return text

    def write_ref_element(self, key, text, sub_indent, source_line=None):
        """ Render a single reference element """
        # Use an empty first line if key is too long
        min_spacing = 2
        if len(key) + min_spacing > sub_indent:
            self.write_text(key, indent=3, leading_blankline=True, wrap_urls=False, fix_sentence_endings=False, source_line=source_line)
            self.write_text(text, indent=3 + sub_indent, wrap_urls=False, fix_sentence_endings=False, source_line=source_line)
        else:
            # Fill space to sub_indent in the bullet
            self.write_text(text, indent=3, bullet=key.ljust(sub_indent), \
                     sub_indent=sub_indent, leading_blankline=True, wrap_urls=False, fix_sentence_endings=False, source_line=source_line)
    
    def _combine_inline_elements(self, elements):
        """ Shared function for <t> and <c> elements
        
            Aggregates all the rendered text of the following elements:
                - xref
                - eref
                - iref
                - cref
                - spanx
            
            Plus their tails.  If an element is encountered that isn't one
            of these (such as a list, figure, etc) then the function
            returns.
            
            This function takes a list of elements as its argument.
            
            This function returns TWO arguments, the aggregated text, and 
            a list containing the rest of the elements that were not processed,
            so that the calling function can deal with them.
        """
        line = ['']
        for i, element in enumerate(elements):
            if element.tag not in self.inline_tags:
                # Not an inline element, exit
                return ''.join(line), elements[i:]

            if element.tag == 'xref':
                line.append(self._expand_xref(element))
            elif element.tag == 'eref':
                if element.text and len(element.text)>0:
                    line.append(element.text.strip())
                    self.eref_count += 1
                    line.append(' [' + str(self.eref_count) + ']')
                    if self.indexmode:
                        self.eref_list.append([self.eref_count, element])
                else:
                    line.append('<' + element.attrib['target'].strip() + '>')
            elif element.tag == 'iref':
                self._add_iref_to_index(element)
            elif element.tag == 'cref':
                if element.pis['comments'] == 'yes':                
                    # Render if processing instruction is enabled
                    self.cref_counter += 1
                    anchor = element.attrib.get('anchor', None)
                    if anchor is None:
                        anchor = 'CREF' + str(self.cref_counter)
                        element.attrib['anchor'] = anchor
                    self._indexCref(self.cref_counter, anchor)
                    if element.pis['inline'] == 'yes':
                        if anchor:
                            anchor = anchor + ': '
                        source = element.attrib.get('source', "")
                        if source:
                            source = " --" + source
                        if element.text:
                            line.append('[[' + anchor + element.text + source + ']]')
                    else:
                        line.append('[' + anchor + ']')
                        self.cref_list.append(element)
            elif element.tag == 'spanx':
                style = element.attrib.get('style', 'emph')
                edgechar = '_'  # default to emph because the spanx element exists
                if style == 'emph':
                    edgechar = '_'
                elif style == 'strong':
                    edgechar = '*'
                elif style == 'verb':
                    edgechar = '"'
                text = ''
                if element.text:
                    text = element.text
                line.append(edgechar + text + edgechar)
            elif element.tag == 'u':
                line.append(self._expand_u(element))
            else:
                xml2rfc.log.error("Found unexpected inline element: <%s>" % element.tag)

            # Add tail text before next element
            if element.tail:
                line.append(element.tail)

            # Go to next sibling
            element = element.getnext()

        # Went through all elements, return text with an empty list
        return ''.join(line), []
            
    def _check_long_lines(self, buf_line, source_line):
        long_lines = [ (num, line) for num, line in enumerate(self.buf[buf_line:]) if (len(line) > self.width) ]
        for num, line in long_lines:
            if source_line:
                xml2rfc.log.warn("Output line (from source around line %s) is %s characters; longer than %s.  Excess characters: '%s':\n  '%s'\n"
                    % (source_line+num, len(line), self.width, line[self.width:], line))
            else:
                xml2rfc.log.warn("Output line (from source around line %s) is %s characters; longer than %s.  Excess characters: '%s':\n  '%s'\n"
                    % (buf_line+num, len(line), self.width, line[self.width:], line))

    # ---------------------------------------------------------
    # Base writer overrides
    # ---------------------------------------------------------

    def insert_toc(self):
        """ Marks buffer position for post-writing table of contents """
        self.toc_marker = len(self.buf)
        
    def insert_iref_index(self):
        """ Marks buffer position for post-writing index """
        self.iref_marker = len(self.buf)

    def write_raw(self, text, indent=3, align='left', blanklines=0, \
                  delimiter=None, leading_blankline=True, source_line=None):
        """ Writes a raw stream of characters, preserving space and breaks """
        
        if text:
            if leading_blankline:
                # Start with a newline
                self._lb()
            # Delimiter?
            if delimiter:
                self.buf.append(delimiter)
            # Additional blank lines?
            self.buf.extend([''] * blanklines)
            start_line = len(self.buf)
            # Format the input
            if "\t" in text:
                xml2rfc.log.warn("Text %scontains tab characters.  These will be expanded, assuming a tab-size of 8." %
                    (("around line %s "%source_line) if source_line else ""))
            lines = [line.rstrip() for line in text.expandtabs().split('\n')]
            # Outdent if it helps anything
            longest_line = max(len(line.rstrip()) for line in lines)
            if (longest_line > self.width-indent):
                new_indent = max(self.width - longest_line, 0)
                if not self.indexmode:
                    xml2rfc.log.warn('artwork outdented %s characters to avoid overrunning right margin around input line %s)' % (indent - new_indent, source_line if source_line else 0))
                indent = new_indent

            # Trim first and last lines if they are blank, whitespace is handled
            # by the `blanklines` and `delimiter` arguments
            if len(lines) > 1:
                if lines[0] == '':
                    lines.pop(0)
                if lines[-1] == '':
                    lines.pop(-1)
            if align == 'center':
                # Find the longest line, and use that as a fixed center.
                center_indent = indent + ((self.width - indent - longest_line) // 2)
                indent_str = center_indent > indent and ' ' * center_indent or \
                                                        ' ' * indent
                for line in lines:
                    self.buf.append(indent_str + line.rstrip())
            elif align == 'right':
                indent_str = ' ' * (self.width - longest_line)
                for line in lines:
                    self.buf.append(indent_str + line.rstrip())
            else:  # align == left
                indent_str = ' ' * indent
                for line in lines:
                    self.buf.append(indent_str + line.rstrip())
            # Additional blank lines?
            self.buf.extend([''] * blanklines)
            # Delimiter?
            if delimiter:
                self.buf.append(delimiter)
            if not self.indexmode:
                self._check_long_lines(start_line, source_line)

    def write_label(self, text, type='figure', source_line=None):
        """ Writes a centered label """
        self.write_text(text, indent=3, align='center', leading_blankline=True, source_line=source_line)

    def write_title(self, title, docName=None, source_line=None):
        """ Write the document title and (optional) name """
        self.write_text(title, leading_blankline=True, align='center', source_line=source_line)
        if docName and not self.rfcnumber:
            self.write_text(docName, align='center')

    def write_heading(self, text, bullet='', autoAnchor=None, anchor=None, \
                      level=1):
        """ Write a generic header """
        if bullet:
            bullet += '  '
        self.write_text(text, bullet=bullet, indent=0, leading_blankline=True)

    def write_paragraph(self, text, align='left', autoAnchor=None):
        """ Write a generic paragraph of text.  Used for boilerplate. """
        text = xml2rfc.utils.urlkeep(text)
        self.write_text(text, indent=3, align=align, leading_blankline=True)

    def write_t_rec(self, t, indent=3, sub_indent=0, bullet='',
                     autoAnchor=None, align='left', level=0, leading_blankline=True):
        """ Recursively writes a <t> element """
        # Grab any initial text in <t>
        current_text = t.text or ''
        if bullet and current_text == '':
            current_text = ' '
        
        # Render child elements
        remainder = t.getchildren()
        while len(remainder) > 0 or current_text:
            # Process any inline elements
            inline_text, remainder = self._combine_inline_elements(remainder)
            current_text += inline_text
            if (current_text and not current_text.isspace()) or bullet:
                # Attempt to write a paragraph of inline text
                self.write_text(current_text, indent=indent,
                                leading_blankline=leading_blankline,
                                sub_indent=sub_indent, bullet=bullet,
                                edit=True, align=align, source_line=t.sourceline)
            # Clear text
            current_text = ''

            # Handle paragraph-based elements (list, figure, vspace)
            if len(remainder) > 0:
                # Get front element
                element = remainder.pop(0)

                if element.tag == 'list': 
                    if sub_indent > 0:
                        new_indent = sub_indent + indent
                    else:
                        new_indent = len(bullet) + indent
                    # Call sibling function to construct list
                    self.write_list(element, indent=new_indent, level=level)
                    # Auto-break for tail paragraph
                    leading_blankline = True
                    bullet = ''

                elif element.tag == 'figure':
                    self.write_figure(element)
                    # Auto-break for tail paragraph
                    leading_blankline = True
                    bullet = ''

                elif element.tag == 'vspace':
                    # Insert `blankLines` blank lines into document
                    self._vspace(num=int(element.attrib.get('blankLines',
                                         self.defaults['vspace_blanklines'])))
                    # Don't auto-break for tail paragraph
                    leading_blankline = False
                    # Keep indentation
                    bullet = ' ' * sub_indent

                # Set tail of element as input text of next paragraph
                if element.tail:
                    current_text = element.tail
                    

    def write_top(self, left_header, right_header):
        """ Combines left and right lists to write a document heading """
        # Begin with a blank line on the first page.  We'll add additional
        # blank lines at the top later, but those won't be counted as part
        # of the page linecount.
        #self._lb(num=3)
        self._lb()
        heading = []
        for i in range(max(len(left_header), len(right_header))):
            if i < len(left_header):
                left = left_header[i]
            else:
                left = ''
            if i < len(right_header):
                right = right_header[i]
            else:
                right = ''
            heading.append(xml2rfc.utils.justify_inline(left, '', right, \
                                                        self.width))
        self.write_raw('\n'.join(heading), align='left', indent=0,
                        leading_blankline=False, source_line=None)
        # Extra blank line underneath top block
        self._lb()

    def write_address_card(self, author):
        """ Writes a simple address card with no line breaks """
        lines = []
        if 'role' in author.attrib:
            lines.append("%s (%s)" % (author.attrib.get('fullname', ''),
                                      author.attrib.get('role', '')))
        else:
            lines.append(author.attrib.get('fullname', ''))
        organization = author.find('organization')
        if organization is not None and organization.text:
            lines.append(organization.text.strip())
        address = author.find('address')
        if address is not None:
            postal = address.find('postal')
            if postal is not None:
                for street in postal.findall('street'):
                    if street.text:
                        lines.append(street.text)
                cityline = []
                city = postal.find('city')
                if city is not None and city.text:
                    cityline.append(city.text)
                region = postal.find('region')
                if region is not None and region.text:
                    if len(cityline) > 0: cityline.append(', ');
                    cityline.append(region.text)
                code = postal.find('code')
                if code is not None and code.text:
                    if len(cityline) > 0: cityline.append('  ');
                    cityline.append(code.text)
                if len(cityline) > 0:
                    lines.append(''.join(cityline))
                country = postal.find('country')
                if country is not None and country.text:
                    lines.append(country.text)
            lines.append('')
            phone = address.find('phone')
            if phone is not None and phone.text:
                lines.append('Phone: ' + phone.text)
            facsimile = address.find('facsimile')
            if facsimile is not None and facsimile.text:
                lines.append('Fax:   ' + facsimile.text)
            email = address.find('email')
            if email is not None and email.text:
                label = self.pis['rfcedstyle'] == 'yes' and 'EMail' or 'Email'
                lines.append('%s: %s' % (label, email.text))
            uri = address.find('uri')
            if uri is not None and uri.text:
                lines.append('URI:   ' + uri.text)
        self.write_raw('\n'.join(lines), indent=self.margin)
        self._lb()

    def write_reference_list(self, list):
        """ Writes a formatted list of <reference> elements """
        refdict = {}
        annotationdict = {}
        refkeys = []
        refsource = {}
        # [surname, initial.,] "title", (STD), (BCP), (RFC), (Month) Year.
        i = 0
        for i, ref in enumerate(list.findall('reference')):
            refstring = []
            authors = ref.findall('front/author')
            author_string = self._format_author_string(authors)
            refstring.append(author_string)
            if len(author_string):
                refstring.append(', ')
            title = ref.find('front/title')
            if title is not None and title.text:
                if ref.attrib.get("quote-title", "true") == "true": # attribute default value: yes
                    refstring.append('"' + title.text.strip() + '"')
                else:
                    refstring.append(title.text.strip())
            else:
                xml2rfc.log.warn('No title specified in reference',
                                 ref.attrib.get('anchor', ''))
            for seriesInfo in ref.findall('seriesInfo'):
                if seriesInfo.attrib['name'] == "Internet-Draft":
                    refstring.append(', '+seriesInfo.attrib['value'] + ' (work in progress)')
                else:
                    refstring.append(', '+seriesInfo.attrib['name'] + u'\u00A0' +
                                     seriesInfo.attrib['value'].replace('/', '/' + u'\uE060'))
            date = ref.find('front/date')
            if date is not None:
                month = date.attrib.get('month', '')
                year = date.attrib.get('year', '')
                if month or year:
                    if month:
                        month += ' '
                    refstring.append(', '+month + year)
            # Target?
            target = ref.attrib.get('target')
            if target:
                refstring.append(', <' + target + '>')
            refstring.append('.')
            annotation = ref.find('annotation')
            # Use anchor or num depending on PI
            if self.pis['symrefs'] == 'yes':
                key = ref.attrib.get('anchor', str(i + self.ref_start))
            else:
                key = str(i + self.ref_start)
            refdict[key] = ''.join(refstring)
            refsource[key] = ref.sourceline
            refkeys.append(key)
            # Add annotation if it exists to a separate dict
            if annotation is not None and annotation.text:
                # Render annotation as a separate paragraph
                remainder = annotation.getchildren()
                annotationdict[key] = annotation.text
                if len(remainder) > 0:
                    inline_txt, remainder = self._combine_inline_elements(remainder)
                    annotationdict[key] += inline_txt
        self.ref_start += i + 1
        # Don't sort if we're doing numbered refs; they are already in
        # numeric order, and if we sort, they will be sorted alphabetically,
        # rather than numerically ... ( i.e., [10], [11], ... [19], [1], ... )
        if self.pis['sortrefs'] == 'yes' and self.pis['symrefs'] == 'yes' :
            refkeys.sort(key=str.lower)
        # Hard coded indentation amount
        refindent = 11
        for key in refkeys:
            reflabel = '['+key+']'
            self.write_ref_element(reflabel, refdict[key], refindent, source_line=refsource[key])
            # Render annotation as a separate paragraph
            if key in annotationdict:
                self.write_text(annotationdict[key], indent=refindent + 3,
                                 leading_blankline=True, source_line=refsource[key])

    def write_erefs(self, refs_counter, refs_subsection):
        bullet=str(refs_counter) + '.' + str(refs_subsection)
        self.write_heading("URIs", bullet=bullet +'.',
                           level = 2, autoAnchor='rfc.references.' + str(refs_subsection))
        for i, element in self.eref_list:
            self.write_text(element.attrib['target'], bullet='['+str(i)+'] ',
                            leading_blankline=True, indent=3)

    def write_crefs(self):
        """ If we have any of these because the are not inline, then
            write them out here
        """
        if self.cref_list:
            self.write_heading('Editorial Comments')
            for cref in self.cref_list:
                body = cref.text
                source = cref.attrib.get('source', None)
                if source:
                    body = source + ': ' + body
                self.write_text(body, bullet='['+cref.attrib['anchor']+'] ',
                                leading_blankline=True)

    def draw_table(self, table, table_num=None):
        # First construct a 2d matrix from the table
        matrix = []
        matrix.append([])
        row = 0
        column_aligns = []
        ttcol_width_attrs = []
        style = table.attrib.get('style', self.defaults['table_style'])

        # Extract information from the ttcol elements about each column
        for ttcol in table.findall('ttcol'):
            column_aligns.append(ttcol.attrib.get('align', self.defaults['ttcol_align']))
            ttcol_width_attrs.append(ttcol.attrib.get('width', ''))

            text = ttcol.text or ''
            if len(ttcol) > 0:
                # <c> has children, render their text and add to line
                inline_text, null = \
                    self._combine_inline_elements(ttcol.getchildren())
                text += inline_text
            matrix[row].append(text)

        num_columns = len(matrix[0])

        # Get the text for each cell into the matrix
        #    Insert appropriate blank lines for filler at this time.
        for i, cell in enumerate(table.findall('c')):
            if i % num_columns == 0:
                # Insert blank row if PI 'compact' is 'no'
                if table.pis["compact"] == "no":
                    if (style == 'none' and row == 0) or (style in ['headers', 'full'] and row != 0):
                        row += 1
                        matrix.append(['']*num_columns)
                        pass
                row += 1
                matrix.append([])


            text = cell.text or ''
            if len(cell) > 0:
                # <c> has children, render their text and add to line
                inline_text, null = \
                    self._combine_inline_elements(cell.getchildren())
                text += inline_text
            matrix[row].append(text)

        # Get table style and determine maximum width of table
        if style in ['none', 'headers']:
            table_max_chars = self.width - 3 - num_columns + 1
        else:
            table_max_chars = self.width - 3 - 3 * num_columns - 1  # indent+border

        if table_max_chars < 0:
            xml2rfc.log.error("too many <ttcol>s to even build <texttable> frame")

        # Find the longest line and longest word in each column
        longest_lines = [0] * num_columns
        longest_words = [0] * num_columns
        for col in range(num_columns):
            for row in matrix:
                if col < len(row) and len(row[col]) > 0:  # Column exists
                    # Longest line
                    if self._length(row[col]) > longest_lines[col]:
                        longest_lines[col] = self._length(row[col])
                    # Longest word
                    if row[col].strip():
                        word = max(xml2rfc.utils.ascii_split(row[col]), key=self._length)
                        if self._length(word) > longest_words[col]:
                            longest_words[col] = self._length(word)
        
        min_width = sum(longest_words)
        
        # Assume one character per non-empty column - are there too many columns?
        inf = []
        for col in longest_words:
            col = (col != 0)
            inf.append(col)
        if table_max_chars < sum(inf):
            xml2rfc.log.error("too many non-empty columns in <texttable> to fit in page width")

        # Translate width specs given available space
        cols = []
        rel_total = 0
        ttcol_width_unspec = []

        for col in ttcol_width_attrs:
            #  what to do with the unspecified ones is touchy.
            u = 0
            if not col:
                x = longest_words[len(cols)]
                u = 1
            elif col == '*':
                x = -1
                rel_total += x
            else:
                m = re.match("^0*([0-9]+)(|em|[%*])$", col)
                if not m:
                    xml2rfc.log.warn("Width attribute not in correct format: %s" & col)
                    x = longest_words[len(cols)]
                    u = 1
                if m.group(2) == 'em':
                    x = int(m.group(1));
                elif m.group(2) == '%':
                    x = (int(m.group(1)) * table_max_chars + 99)//100
                elif m.group(2) == '*':
                    x = -int(m.group(1))
                    rel_total += x

            cols.append(x)
            ttcol_width_unspec.append(u)

        ttcol_width_attrs = cols

        # Compute final cell widths ( and their totals )

        width = table_max_chars
        if table_max_chars == sum(longest_words):
            #  Use the minimum widths if that is all we have
            ttcol_widths = longest_words
        elif table_max_chars == sum(inf):
            #  ??????????????
            ttcol_widths = inf
        elif table_max_chars < sum(longest_words):
            if not self.indexmode:
                xml2rfc.log.warn("so many <ttcol>s in <texttable> that some words need to be split near line %s" % table.sourceline)

            excess = sum(longest_words) - table_max_chars

            # First, try the mins that are greater than what they requested.

            rel_total = -rel_total
            tol = []
            fnum  = 1000 if (rel_total > 0) else 0
            fden  = 1
            for s, n in zip(ttcol_width_attrs, longest_words):
                t = 0
                if s >= 0:
                    s += (s == 0)
                    y = n - s
                    if y > 0:
                        t = y
                elif fnum > 0:
                    # Prepare in case we need to try the relative widths.
                    n -= 1
                    s = -s
                    # We keep the smallest acceptable factor to all as a mixed number.
                    if n <= 0:
                        fnum = 0
                    elif n * fden < s * fnum:
                        fnum = n
                        fden = s

                tol.append(t)

            excess, ttcol_widths = self.shave_cols_excess(longest_words, excess, tol, sum(tol))

            # Second, try the unspecified widths.
            if excess > 0:
                tol = []
                for w,u in zip(ttcol_widths, ttcol_width_unspec):
                    t = (w-1) if (u and w > 0) else 0
                    tol.append(t)

                excess, ttcol_widths = self.shave_cols_excess(ttcol_widths, excess, tol, sum(tol))

            # Third, try the relative widths (to shrink, not expand).
            if excess > 0 and fnum > 0:
                if excess * fden < rel_total * fnum:
                    fnum = excess
                    fden = rel_total

                tol = []
                for s in ttcol_width_attrs:
                    t = ((fnum * -(s)) // fden) if (s < 0) else 0
                    tol.append(t)

                excess, ttcol_widths = self.shave_cols_excess(ttcol_widths, excess, tol, sum(tol))

            # Finally, try anything that's left, leaving as little as one column.
            # Start with largest.

            p = 1
            for w in ttcol_widths:
                if w > p:
                    p = w
            
            while excess > 0:
                if p == 1:
                    xml2rfc.log.error("bug in reducing <texttable> column widths")

                # Find second largest.
                q = 1
                for w in ttcol_widths:
                    if w > q and w < p:
                        q = w

                tol = []
                for w in ttcol_widths:
                    t = (w - q) if (w > q) else 0
                    tol.append(t)

                excess, ttcol_widths = self.shave_cols_excess(ttcol_widths, excess, tol, sum(tol))

                p = q

        else:
           #  There is room beyond the minimum space
           #  First, try and give everyone what they want or need
           #   longest_words + ttcol_width_attrs + longest_lines(all)
           ttcol_widths = []
           cols_width = 0

           #   longest_words + ttcol_width_attrs
           mnsa = []

           #   longest_words + ttcol_width_attrs + longest_lines(rel)
           mnsb = []

           #   longest_words + ttcol_width_attrs + longest_lines(rel + ttcol_width_unspec)
           mnsc = []
       
           for m, n, s, u in zip(longest_lines, longest_words, ttcol_width_attrs, ttcol_width_unspec):
                t = n
                if (t < s):
                    t = s
                mnsa.append( t )

                if (s < 0):
                    t = m
                mnsb.append( t )

                if u:
                    t = m
                mnsc.append(t)

                if (t < m):
                    t = m
                ttcol_widths.append(t)

           cols_width = sum(ttcol_widths)

           #  In decreasing order by width
           if table_max_chars >= cols_width:
               width = cols_width

           elif table_max_chars > sum(mnsc):
               # partially expand columns with greater maximal width from mnsc
               target = table_max_chars - sum(mnsc)
               target, ttcol_widths = self.expand_cols_by_height(mnsc, target, ttcol_widths)

           elif table_max_chars == sum(mnsc):
                ttcol_widths = mnsc

           elif table_max_chars > sum(mnsb):
               # Partially expand columns with unspecified width from mnsb
               target = table_max_chars - sum(mnsb)
               target, ttcol_widths = self.expand_cols_by_height(mnsb, target, ttcol_widths)

           elif table_max_chars == sum(mnsb):
                ttcol_widths = mnsb

           elif table_max_chars > sum(mnsa):
                #Partially expand columns with relative width from mnsa.
            
                ttcol_widths = mnsa
                target = table_max_chars - sum(mnsa)
                while target > 0:
                   rel = []
                   num_rel = 0
                   rel_total = 0
                   fnum = 1000
                   fden = 1
                   for m,s,w in zip(longest_lines, ttcol_width_attrs, ttcol_widths):
                       x = 0
                       if s < 0:
                           d = m - w
                           if d > 0:
                               x = -s
                               if d * fden < x * fnum:
                                   fnum = d
                                   fden = x
                               num_rel += 1;
                               rel_total += x
                       rel.append(x)
                       
                   if target * fden < rel_total * fnum:
                       fnum = target
                       fden = rel_total

                   plus_width = 0
                   if (num_rel > 0) and (fnum > 0):
                       kols = []
                       for r,w in zip(rel, ttcol_widths):
                           p = r*fnum//fden
                           kols.append(w + p)
                           plus_width += p

                   if plus_width > 0:
                       ttcol_widths = kols
                       target += -plus_width

                   elif num_rel > 0:
                       dd = target
                       kols = []
                       for r,w in zip(rel, ttcol_widths):
                           if r > 0:
                              # prefer expanding the rightmost ones
                              if dd >= num_rel:
                                  w += 1
                                  target -= 1
                              dd += 1
                           kols.append(w)
                       ttcol_widths = kols

                   else:
                       xml2rfc.log.error("bug in expanding <texttable> columns with relative widths")

           elif table_max_chars == sum(mnsa):
               ttcol_widths = mnsa
           else:
               # Partially expand comns with greater specified width from longest_words
               target = table_max_chars - min_width
               plus = []
               for a,b in zip(mnsa, longest_words):
                   plus.append(a-b)
               target, ttcol_widths = self.expand_cols(longest_words, target, plus)

        # Ensure we don't have any zero-width columns; it breaks textwrap
        ttcol_widths = [ k or 1 for k in ttcol_widths ] 
        # Now construct the cells using textwrap against ttcol_widths
        cell_lines = [
            [ textwrap.wrap(cell, ttcol_widths[j]) or [''] for j, cell in enumerate(matrix[i]) ]
            for i in range(0, len(matrix))
        ]

        output = []
        # Create the border
        if style == 'none':
            pass
        elif style == 'headers':
            borderstring = []
            for i in range(num_columns):
                borderstring.append('-' * ttcol_widths[i])
                borderstring.append(' ')
        else:
            borderstring = ['+']
            for i in range(num_columns):
                borderstring.append('-' * (ttcol_widths[i] + 2))
                borderstring.append('+')
            output.append(''.join(borderstring))

        # Draw the table
        for i, cell_line in enumerate(cell_lines):
            if i==0 and cell_line == [['']]*num_columns:
                if style in ['headers']:
                    # This is the header row, append the header decoration
                    output.append(''.join(borderstring))
                continue
            # produce as many output rows as the number of wrapped
            # text lines in the cell with most lines, but at least 1
            for row in range(0, max(map(len, cell_line))):
                if style == 'headers' or style == 'none':
                    line = ['']
                else:
                    line = ['|']
                for col, cell in enumerate(cell_line):
                    align = column_aligns[col]
                    width = ttcol_widths[col]
                    if row < len(cell):
                        # width = width + len(cell[row]) - self._length(cell[row])
                        if align == 'center':
                            text = cell[row].center(width)
                        elif align == 'right':
                            text = cell[row].rjust(width)
                        else:  # align == left
                            text = cell[row].ljust(width)
                        if style == 'headers' or style == 'none':
                            line.append(text)
                            line.append(' ')
                        else:
                            line.append(' ')
                            line.append(text)
                            line.append(' |')
                    else:
                        if style == 'headers' or style == 'none':
                            line.append(' ' * (ttcol_widths[col] + 1))
                        else:
                            line.append(' ' * (ttcol_widths[col] + 2) + '|')
                output.append(''.join(line))
            if style in ['headers', 'full']:
                if i == 0:
                    # This is the header row, append the header decoration
                    output.append(''.join(borderstring))
            if style in ['full']:
                if i == len(cell_lines)-1:
                    output.append(''.join(borderstring))
            if style in ['all']:
                output.append(''.join(borderstring))


        # Finally, write the table to the buffer with proper alignment
        align = table.attrib.get('align', self.defaults['table_align'])
        self.write_raw('\n'.join(output), align=align, indent=self.margin,
                        source_line=table.sourceline)


    def expand_cols_by_height(self, cols, target, bound):
        # cols - current column width
        # target - expand by this amount 
        # bound - this is the lower boundary of the column width

        while target > 0:
            # Compute the current width and boundary width for those where current is less than boundary
            m_width = 0
            w_width = 0
            for m,w in zip(bound, cols):
                if (m > w):
                   # Only cover those columns below the bound.
                   m_width += m
                   w_width += w

            #  This mixed number is the average height we could achieve.

            fnum = m_width
            fden = w_width + target
            plus_num = 0
            plus = []
            plus_width = 0
            kols = []
            for m,w in zip(bound, cols):
                p = 0
                k = w
                if m > w:
                    x = m * fden
                    y = w * fnum
                    # Seek columns of above average height.
                    if x >= y:
                        # Nearly reduce height to average by augmenting width
                        p = 1
                        k = x // (fnum + fden - 1)
                        if k < w:
                            # but no smaller than before.
                            k = w
                        if k > w + target:
                            k = w + target
                        q = k - w
                        plus_num += 1
                        plus_width += q

                plus.append(p)
                kols.append(k)

            if plus_width > 0:
               cols = kols
               target += (-plus_width)

            elif plus_num > 0:
                # No progress due to integer division.  Make some.
                dd = target
                cols = []
                for p,w in zip(plus, kols):
                    if p > 0:
                        # Prefer expanding the rightmost ones.
                        if dd >= plus_num:
                            w += 1
                            target -= 1
                        dd += 1
                    cols.append(w)

            else:
                # Should not happen, but just in case we were given a bad target
                cols = cols
                break;

        return [target, cols]

    def shave_cols_excess (self, cols, excess, give, give_width=-1):
        if give_width < 0:
            give_width =sum(give)

        if excess <= 0 or give_width <= 0:
            # Nothing to do.
            pass
        elif excess >= give_width:
            # Take everything; it may still not be enough.
            plus = []
            for a,b in zip(cols, give):
                plus.append(a-b)
            cols = plus
            excess += -(give_width)
        else:
            # Take only what we need, in relative proportion.
            dd = 0
            kols = []
            for c, g in zip(cols, give):
                if g > 0:
                    d =(g * excess) // give_width
                    c += -(d)
                    dd += d
                kols.append( c )
            excess += -(dd)
    
            # Do the ones we missed because of integer division.
            if excess > 0:
                cols = []
                for k, g in zip(kols, give):
                    # Prefer shaving the leftmost ones.
                    if excess > 0 and g > 0:
                        k += -1
                        excess += -1
                    cols.append( k )
            else:
                cols = kols
    
        return [excess, cols]

    def expand_cols(self, cols, target, plus, plus_width=-1):
        if plus_width < 0:
            plus_width =sum(plus)

        if target <= 0 or plus_width <= 0:
            # Nothing to do.
            pass
        elif target >= plus_width:
            # Take everything; it may still not be enough.
            tmp = []
            for a,b in zip(cols, plus):
                tmp.append(a+b)
            cols = tmp
            target += -(plus_width)
        else:
            # Take only what we need, in relative proportion.
            dd = 0
            dr = 0
            kols = []
            for c,p in zip(cols, plus):
                if p > 0:
                    d = (p * target) // plus_width
                    c += d
                    dd += d
                    dr += 1
                kols.append(c)

            target += -(dd)

            # Do the ones we missed because of integer division.
            if target > 0:
                dd = target
                cols = []
                for k, p in zip(kols, plus):
                    if p > 0:
                        # Prefer expanding the rightmost ones.
                        if dd >= dr:
                            k += 1
                            target -= 1
                        dd += 1
                    cols.append( k )
            else:
                cols = kols

        return [target, cols]

    def insert_anchor(self, text):
        # No anchors for text
        pass

    def pre_indexing(self):
        pass

    def pre_rendering(self):
        # Discard buffer from indexing pass
        self.buf = []
        
        # Reset document counters from indexing pass
        self.list_counters = {}
        self.edit_counter = 0   # Counter for edit marks
        self.cref_counter = 0
        self.cref_list = []

    def post_rendering(self):
        # Insert the TOC and IREF into the main buffer
        self.output = self.buf[:self.toc_marker] + \
                      self.write_toc()
        if self.iref_marker > 0:
            self.output += self.buf[self.toc_marker:self.iref_marker] + \
                      self.write_iref_index() + \
                      self.buf[self.iref_marker:]
        else:
            self.output += self.buf[self.toc_marker:]

    def post_process_lines(self, lines):
        output = []
        for line in lines:
            if isinstance(line, type(u'')):
                # line = line.replace(u"\u8209", '-')
                line = line.replace(u'\u00A0', u' ')
                line = line.replace(u'\u2011', u'-')
                line = line.replace(u'\u200B', u'')
                line = line.replace(u'\u2028', u' ')
                line = line.replace(u'\u2060', u'')
                assert line == line.replace(u'\uE060', u'')
            output.append(line);
        return output

    def write_to_file(self, file):
        """ Writes the buffer to the specified file """
        # write initial blank lines, not counted against page size...
        if self.draft:
            file.write("\n"*3)
        else:
            file.write("\n"*5)
        for line in self.post_process_lines(self.output):
            file.write(line.rstrip(u" \t"))
            file.write("\n")



# Copyright The IETF Trust 2018, All Rights Reserved
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, division

# ----------------------------------------------------------------------
# Unicode operations

import re
import unicodedata

try:
    from xml2rfc import debug
    assert debug
except ImportError:
    pass

from wcwidth import wcswidth

unicode_content_tags = set([
    'artwork',
    'city',
    'cityarea',
    'code',
    'country',
    'email',
    'extaddr',
    'organization',
    'pobox',
    'postalLine',
    'refcontent',
    'region',
    'sortingcode',
    'sourcecode',
    'street',
    'title',
    'u',
])

# Attribute values should not contain unicode, with some exceptions
unicode_attributes = {
    ('author', 'fullname'),
    ('author', 'surname'),
    ('author', 'initials'),
    ('contact', 'fullname'),
    ('contact', 'surname'),
    ('contact', 'initials'),
    ('organization', 'abbrev'),
    ('seriesInfo', 'name'),
    ('seriesInfo', 'value'),
    ('xref', 'derivedContent'),
}

latinscript_attributes = {
    ('author', 'asciiFullname'),
    ('author', 'asciiSurname'),
    ('author', 'asciiInitials'),
    ('contact', 'asciiFullname'),
    ('contact', 'asciiSurname'),
    ('contact', 'asciiInitials'),
}

# These unicode tags don't require an ascii attribute
bare_unicode_tags = set([
    'artwork',
    'refcontent',
    'sourcecode',
    'u',
    ])

bare_latin_tags = set([
    'city',
    'cityarea',
    'code',
    'country',
    'email',
    'extaddr',
    'organization',
    'pobox',
    'postalLine',
    'region',
    'sortingcode',
    'street',
    'title',
    ])

def is_svg(e):
    '''
    Returns true if an element is a SVG element
    '''

    if str(e.tag).startswith('{http://www.w3.org/2000/svg}'):
        return True
    else:
        return False

for t in bare_unicode_tags:
    assert t in unicode_content_tags

def expand_unicode_element(e, bare=False):
    if not e.text:
        return ''
    format = e.get('format', 'lit-name-num')
    size = len(e.text)
    permitted = ('name', 'num', 'char', 'lit', 'ascii')
    required  = ('num', )
    #
    if not format:
        raise ValueError('Expected <unicode> to have a format attribute, but found none')
    def num(e):
        return ' '.join('U+%04X'%ord(c) for c in e.text)
    def lit(e):
        return '"%s"' % e.text
    def char(e):
        return '%s' % e.text
    def name(e):
        try:
            names = []
            for c in e.text:
                if isinstance(c, bytes):
                    c = c.decode('latin1')
                names.append(unicodedata.name(c, 'U+%04x'%ord(c)))
            return ', '.join(names)
        except ValueError as exc:
            raise ValueError('When looking for the name of "%s", got: %s' % (c, exc))
    def ascii(e):
        ascii = e.get('ascii')
        if not ascii:
            raise ValueError('Expected an ascii value when using "ascii" in the <unicode> expansion, but found none')
        return '"%s"' % ascii
    #
    if size == 0:
        return ''
    #
    simple_regex = '^{part}(-{part})*$'.format(part='(%s)' % '|'.join(permitted))
    full = None
    if re.match(simple_regex, format):
        full = False
        val = format
        parts = format.split('-')
    else:
        full = True
        val = format
        parts = re.findall('{([a-z]+)}', format)
    if len(parts) == 0 or not all(p for p in permitted):
        raise ValueError('Expected <unicode> expansion parts to be one or more of %s, but found format="%s"' % (', '.join(permitted), val))
    if not any(p in parts for p in required):
        if len(required) > 1:
            raise ValueError('Expected <unicode> expansion parts to include at least one of %s, but found format="%s"' % (', '.join(['"%s"'%r for r in required]), val))
        else:
            raise ValueError('Expected <unicode> expansion parts to include at least "%s", but found format="%s"' % (required[0], val))            
    if len(parts) > 3:
        raise ValueError('Expected up to 3 dash-separated <unicode> expansion parts, but found %d: format="%s"' % (len(parts), format))
    #
    values_list = []
    values_dict = {}
    for p in parts:
        if p in locals():
            func = locals()[p]
        else:
            raise RuntimeError('Internal Error: looked for a <unicode> expansion rendering function %s(), but didn\'t find it' % p)
        value = func(e)
        values_list.append(value)
        values_dict[p] = value
    #
    if full:
        text = format.format(**values_dict)
    else:
        if   len(parts) == 1:
            template = '%s'
        elif len(parts) == 2:
            template = '%s (%s)'
        elif len(parts) == 3:
            template = '%s (%s, %s)'
        else:
            raise ValueError('Did not expect to be asked to render <%s> with format="%s"' % (e.tag, format))
        text = template % tuple(values_list)
    return text

def isascii(u):
    if u is None:
        return True
    if isinstance(u, str):
        t = u+''
        for ch in [ '\u00a0', '\u200B', '\u2011', '\u2028', '\u2060', ]:
            if ch in t:
                t = t.replace(ch, ' ')
        try:
            t.encode('ascii')
            return True
        except UnicodeEncodeError:
            return False
    else:
        return True

def textwidth(u):
    "Length of string, disregarding zero-width code points"
    return wcswidth(u)

def downcode_punctuation(str):
    while True:
        match = re.search(punctuation_re, str)
        if not match:
            return str
        str = re.sub(match.group(0), punctuation[match.group(0)], str)

def downcode(str, replacements=None, use_charrefs=True):
    """

    Replaces Unicode characters that we do not use internally with selected
    replacement strings or with str.encode()'s xmlcharrefreplace string.

    Characters used internally, and stripped before emission:

        '\u00a0'                     # non-breaking whitespace
        '\u2060'                     # word joiner
        '\u200B'                     # zero-width space
        '\u2011'                     # non-breaking hyphen
        '\u2028'                     # line separator
        '\uE060'                     # word joiner

    """
    if not replacements:
        replacements = unicode_replacements

    while True:
        match = re.search(u'([^ -\x7e\u2060\u200B\u00A0\u2011\u2028\uE060\r\n\t])', str)
        if not match:
            return str
        if   match.group(1) in replacements:
            str = re.sub(match.group(1), replacements[match.group(1)], str)
        elif match.group(1) in controlchars:
            str = re.sub(match.group(1), controlchars[match.group(1)], str)
        else:
            entity = match.group(1).encode('ascii', 'xmlcharrefreplace').decode('ascii')
            str = re.sub(match.group(1), entity, str)

unicode_space_replacements = {
    u'\u2002': ' ',
    u'\u2003': ' ',
    u'\u2009': ' ',
}

unicode_dash_replacements = {
    u'\u002d': '-',
    u'\u2010': '-',
    u'\u2013': '-',
    u'\u2014': '-',
    u'\u2212': '-',
}

unicode_quote_replacements = {
    u'\u00b4': "'",
    u'\u2018': "'",
    u'\u2019': "'",
    u'\u201a': "'",
    u'\u201c': '"',
    u'\u201d': '"',
    u'\u201e': '"',
    u'\u2032': "'",
}

punctuation = {
    u'\u2026': '...',
}
punctuation.update(unicode_space_replacements)
punctuation.update(unicode_dash_replacements)
punctuation.update(unicode_quote_replacements)
punctuation_re = re.compile(r'[%s]'%''.join(list(punctuation.keys())))

unicode_replacements = {
    # Unicode code points corresponding to (x)html entities, also in
    # rfc2629-xhtml.ent
    u'\x09': ' ',
    u'\xa0': ' ',
    u'\xa1': '!',
    u'\xa2': '[cents]',
    u'\xa3': 'GBP',
    u'\xa4': '[currency units]',
    u'\xa5': 'JPY',
    u'\xa6': '|',
    u'\xa7': 'S.',
    u'\xa9': '(C)',
    u'\xaa': 'a',
    u'\xab': '<<',
    u'\xac': '[not]',
    u'\xae': '(R)',
    u'\xaf': '_',
    u'\xb0': 'o',
    u'\xb1': '+/-',
    u'\xb2': '^2',
    u'\xb3': '^3',
    u'\xb4': "'",
    u'\xb5': '[micro]',
    u'\xb6': 'P.',
    u'\xb7': '.',
    u'\xb8': ',',
    u'\xb9': '^1',
    u'\xba': 'o',
    u'\xbb': '>>',
    u'\xbc': '1/4',
    u'\xbd': '1/2',
    u'\xbe': '3/4',
    u'\xbf': '?',
    u'\xc0': 'A',
    u'\xc1': 'A',
    u'\xc2': 'A',
    u'\xc3': 'A',
    u'\xc4': 'Ae',
    u'\xc5': 'Ae',
    u'\xc6': 'AE',
    u'\xc7': 'C',
    u'\xc8': 'E',
    u'\xc9': 'E',
    u'\xca': 'E',
    u'\xcb': 'E',
    u'\xcc': 'I',
    u'\xcd': 'I',
    u'\xce': 'I',
    u'\xcf': 'I',
    u'\xd0': '[ETH]',
    u'\xd1': 'N',
    u'\xd2': 'O',
    u'\xd3': 'O',
    u'\xd4': 'O',
    u'\xd5': 'O',
    u'\xd6': 'Oe',
    u'\xd7': 'x',
    u'\xd8': 'Oe',
    u'\xd9': 'U',
    u'\xda': 'U',
    u'\xdb': 'U',
    u'\xdc': 'Ue',
    u'\xdd': 'Y',
    u'\xde': '[THORN]',
    u'\xdf': 'ss',
    u'\xe0': 'a',
    u'\xe1': 'a',
    u'\xe2': 'a',
    u'\xe3': 'a',
    u'\xe4': 'ae',
    u'\xe5': 'ae',
    u'\xe6': 'ae',
    u'\xe7': 'c',
    u'\xe8': 'e',
    u'\xe9': 'e',
    u'\xea': 'e',
    u'\xeb': 'e',
    u'\xec': 'i',
    u'\xed': 'i',
    u'\xee': 'i',
    u'\xef': 'i',
    u'\xf0': '[eth]',
    u'\xf1': 'n',
    u'\xf2': 'o',
    u'\xf3': 'o',
    u'\xf4': 'o',
    u'\xf5': 'o',
    u'\xf6': 'oe',
    u'\xf7': '/',
    u'\xf8': 'oe',
    u'\xf9': 'u',
    u'\xfa': 'u',
    u'\xfb': 'u',
    u'\xfc': 'ue',
    u'\xfd': 'y',
    u'\xfe': '[thorn]',
    u'\xff': 'y',
    u'\u0152': 'OE',
    u'\u0153': 'oe',
    u'\u0161': 's',
    u'\u0178': 'Y',
    u'\u0192': 'f',
    u'\u02dc': '~',
    u'\u2002': ' ',
    u'\u2003': ' ',
    u'\u2009': ' ',
    u'\u2013': '-',
    u'\u2014': u'-\u002D',
    u'\u2018': "'",
    u'\u2019': "'",
    u'\u201a': "'",
    u'\u201c': '"',
    u'\u201d': '"',
    u'\u201e': '"',
    u'\u2020': '*!*',
    u'\u2021': '*!!*',
    u'\u2022': 'o',
    u'\u2026': '...',
    u'\u2030': '[/1000]',
    u'\u2032': "'",
    u'\u2039': '<',
    u'\u203a': '>',
    u'\u2044': '/',
    u'\u20ac': 'EUR',
    u'\u2122': '[TM]',
    u'\u2190': '<-\u002D',
    u'\u2192': '\u002D->',
    u'\u2194': '<->',
    u'\u21d0': '<==',
    u'\u21d2': '==>',
    u'\u21d4': '<=>',
    u'\u2212': '-',
    u'\u2217': '*',
    u'\u2264': '<=',
    u'\u2265': '>=',
    u'\u2329': '<',
    u'\u232a': '>',

    # rfc2629-other.ent
    u'\u0021': '!',
    u'\u0023': '#',
    u'\u0024': '$',
    u'\u0025': '%',
    u'\u0028': '(',
    u'\u0029': ')',
    u'\u002a': '*',
    u'\u002b': '+',
    u'\u002c': ',',
    u'\u002d': '-',
    u'\u002e': '.',
    u'\u002f': '/',
    u'\u003a': ':',
    u'\u003b': ';',
    u'\u003d': '=',
    u'\u003f': '?',
    u'\u0040': '@',
    u'\u005b': '[',
    u'\u005d': ']',
    u'\u005e': '^',
    u'\u005f': '_',
    u'\u0060': '`',
    u'\u007b': '{',
    u'\u007c': '|',
    u'\u007d': '}',
    u'\u017d': 'Z',
    u'\u017e': 'z',
    u'\u2010': '-',
}

controlchars = dict( (str(chr(i)), ' ') for i in range(0, 32) if not i in [ 9, 10, 13 ] )
    



# Copyright The IETF Trust 2025, All Rights Reserved
# -*- coding: utf-8 -*-
import os
import re


class FileAccessError(Exception):
    """File access error"""

    pass


def can_access(options, source, path, access_templates=False):
    # templates can be always accessed
    template_path = os.path.abspath(os.path.join(options.template_dir, path))
    if (
        access_templates
        and template_path.startswith(options.template_dir)
        and os.path.exists(template_path)
    ):
        return True

    # user allowed access?
    if not options.allow_local_file_access:
        raise FileAccessError(
            f"Can not access local file: {path}. Use --allow-local-file-access enable access."
        )

    # does it have shell meta-chars?
    shellmeta = re.compile("[><*[`$|;&(#]")
    if shellmeta.search(path):
        raise FileAccessError(f"Found disallowed shell meta-characters in {path}")

    # file is within the source dir?
    dir = os.path.abspath(os.path.dirname(source))
    path = os.path.abspath(os.path.join(dir, path))
    if not path.startswith(dir):
        raise FileAccessError(
            f"Expected a file located beside or below the .xml source (in {dir}), but found a reference to {path}"
        )

    # does it exists?
    if not os.path.exists(path):
        raise FileAccessError(f"Expected a file at '{path}', but no such file exists")

    return True



# Copyright The IETF Trust 2017, All Rights Reserved
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function

import intervaltree
import os
from codecs import open

try:
    import debug
    debug.debug = True
except ImportError:
    pass

scripts = intervaltree.IntervalTree()

def init():
    # Clocked to less than 0.2 seconds on a 2GHz core 08 Feb 2018
    fn = os.path.join(os.path.dirname(__file__), 'data', 'Scripts.txt')
    with open(fn, encoding='utf-8') as f:
        text = f.read()
    for line in text.splitlines():
        if line.startswith('#'):
            continue
        if not line:
            continue
        line = line.strip()
        data, comment = line.split('#', 1)
        codepoints, script = data.split(';', 1)
        if '..' in codepoints:
            start, stop = [ int(x.strip(), base=16) for x in codepoints.split('..') ]
            scripts[start:stop+1] = script.strip()
        else:
            start = int(codepoints.strip(), base=16)
            scripts[start:start+1] = script.strip()

def get_scripts(text):
    """"Return the unicode scripts used in text.

    Assumes that ascii characters are common in the input.
    Clocked to about 0.5 seconds per million characters
    on a 2Ghz core given input with about 60% ascii.
    """
    scriptset = set()
    for i, c in enumerate(text):
        o = ord(c)
        if i < 100 or o > 127:
            scriptset |= scripts[o]
    return set([ s.data for s in scriptset ])

init()




import os
import sys
import time as timeutils
import inspect

try:
    import syslog
    logger = syslog.syslog
except ImportError:                     # import syslog will fail on Windows boxes
    import logging
    logging.basicConfig(filename='tracker.log',level=logging.INFO)
    logger = logging.info

try:
    from pprint import pformat
except ImportError:
    pformat = lambda x: x

import cProfile
import traceback as tb

# A debug decorator, written by Paul Butler, taken from
# http://paulbutler.org/archives/python-debugging-with-decorators/
# Additional functions and decorator functionality added by
# Henrik Levkowetz

__version__ = "0.16"

increment = 2

debug = False

# Number of times to indent output
# A list is used to force access by reference
_report_indent = [4]
_mark = [ timeutils.time() ]

def set_indent(i):
    _report_indent[0] = i

def trace(fn):                 # renamed from 'report' by henrik 16 Jun 2011
    """Decorator to print information about a function
    call for use while debugging.
    Prints function name, arguments, and call number
    when the function is called. Prints this information
    again along with the return value when the function
    returns.
    """
    def fix(s,n=64):
        import re
        s = re.sub(r'\\t', ' ', s)
        s = re.sub(r'\s+', ' ', s)
        if len(s) > n+3:
            s = s[:n]+"..."
        return s
    def wrap(fn, *params,**kwargs):
        call = wrap.callcount = wrap.callcount + 1

        indent = ' ' * _report_indent[0]
        fr = tb.format_stack()[-3].strip()[4:]                      # call from
        fi, co = [ l.strip() for l in fr.splitlines()[:2] ]         # file info, code
        fu = "%s.%s()" % (fn.__module__, fn.__name__)               # function name
        fc = "%s(%s)" % (fn.__name__, ', '.join(                    # function call
            [fix(repr(a)) for a in params] +
            ["%s = %s" % (a, fix(repr(b))) for a,b in kwargs.items()]
        ))

        if debug:
            sys.stderr.write("\n%s  From %s:\n%s  |  %s\n%s  %s\n%s* %s [#%s]\n" %
                (indent, fi, indent, co, indent, fu, indent, fc, call))
        _report_indent[0] += increment
        mark = timeutils.time()
        ret = fn(*params,**kwargs)
        tau = timeutils.time() - mark
        _report_indent[0] -= increment
        if debug:
            sys.stderr.write("%s  %s | %.3fs [#%s] ==> %s\n" % (indent, fc, tau, call, fix(repr(ret))))

        return ret
    wrap.callcount = 0
    if debug:
        from decorator import decorator
        return decorator(wrap, fn)
    else:
        return fn

def mark():
    def show_entry(e):
        sys.stderr.write(" at %s:L%s %s() %s\n" % e)
    if debug:
        indent = ' ' * (_report_indent[0])
        file, line, func, text = tb.extract_stack(None, 2)[0]
        parts = file.split(os.sep)
        name = os.sep.join(parts[-2:])
        sys.stderr.write("%sMark %s:%s\n" % (indent, name, line))
        _mark[0] = timeutils.time()

def lap(s):
    if debug:
        clk = timeutils.time()
        tau = clk - _mark[0]
        ts = timeutils.strftime("%H:%M:%S", timeutils.localtime(clk))
        say("%s: %.3fs since mark: %s" % (ts, tau, s))

def clock(s):
    if debug:
        lap(s)
        _mark[0] = timeutils.time()

def time(fn):
    """Decorator to print timing information about a function call.
    """
    def wrap(fn, *params,**kwargs):

        indent = ' ' * _report_indent[0]
        fc = "%s.%s()" % (fn.__module__, fn.__name__,)

        mark = timeutils.time()
        ret = fn(*params,**kwargs)
        tau = timeutils.time() - mark
        sys.stderr.write("%s| %s | %.3fs\n" % (indent, fc, tau))

        return ret
    wrap.callcount = 0
    if debug:
        from decorator import decorator
        return decorator(wrap, fn)
    else:
        return fn

def show(name):
    if debug:
        frame = inspect.stack()[1][0]
        value = eval(name, frame.f_globals, frame.f_locals)
        indent = ' ' * (_report_indent[0])
        if sys.version_info.major == 2:
            sys.stderr.write((u"%s%s: '%s'\n" % (indent, name, value)).encode('utf8'))
        else:
            sys.stderr.write((u"%s%s: '%s'\n" % (indent, name, value)))

def log(name):
    if debug:
        frame = inspect.stack()[1][0]
        value = eval(name, frame.f_globals, frame.f_locals)
        indent = ' ' * (_report_indent[0])
        logger("%s%s: %s" % (indent, name, value))

def pprint(name):
    if debug:
        frame = inspect.stack()[1][0]
        value = eval(name, frame.f_globals, frame.f_locals)
        indent = ' ' * (_report_indent[0])
        sys.stderr.write("%s%s:\n" % (indent, name))
        lines = pformat(value).split('\n')
        for line in lines:
            sys.stderr.write("%s %s\n"%(indent, line))

def dir(name):
    if debug:
        name = "dir(%s)" % name
        frame = inspect.stack()[1][0]
        value = eval(name, frame.f_globals, frame.f_locals)
        indent = ' ' * (_report_indent[0])
        sys.stderr.write("%s%s:\n" % (indent, name))
        lines = pformat(value).split('\n')
        for line in lines:
            sys.stderr.write("%s %s\n"%(indent, line))

def type(name):
    if debug:
        name = "type(%s)" % name
        frame = inspect.stack()[1][0]
        value = eval(name, frame.f_globals, frame.f_locals)
        indent = ' ' * (_report_indent[0])
        sys.stderr.write("%s%s: %s\n" % (indent, name, value))
            
def say(s):
    if debug:
        indent = ' ' * (_report_indent[0])
        sys.stderr.write("%s%s\n" % (indent, s))
        sys.stderr.flush()

def profile(fn):
    def wrapper(*args, **kwargs):
        datafn = fn.__name__ + ".profile" # Name the data file sensibly
        prof = cProfile.Profile()
        retval = prof.runcall(fn, *args, **kwargs)
        prof.dump_stats(datafn)
        return retval
    if debug:
        from decorator import decorator
        return decorator(wrapper, fn)
    else:
        return fn
    
def traceback(levels=None):
    if debug:
        indent = ' ' * (_report_indent[0])
        sys.stderr.write("\n%s---- Stack ----\n" % indent)
        if levels:
            start = -2-levels
        else:
            start = None
        for s in tb.format_stack()[start:-1]:
            sys.stderr.write("%s%s" % (indent, s))
        sys.stderr.write("%s---------------\n" % indent)

def show_caller(level=None):
    if debug:
        indent = ' ' * (_report_indent[0])
        if level is None:
            level = -3
        sys.stderr.write("%sCalled from %s\n" % (indent, tb.format_stack()[level].strip()[4:]))

def info(name):
    if debug:
        frame = inspect.stack()[1][0]
        value = eval(name, frame.f_globals, frame.f_locals)
        vtype = eval("type(%s)"%name, frame.f_globals, frame.f_locals)
        indent = ' ' * (_report_indent[0])
        sys.stderr.write("%s%s: '%s' (%s)\n" % (indent, name, value, vtype))



# Copyright The IETF Trust 2017, All Rights Reserved
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function

# The following text entries were taken from
# https://www.ietf.org/ietf-ftp/1id-guidelines.txt

boilerplate_draft_status_of_memo = [
        """<t>
        This Internet-Draft is submitted in full conformance with the
        provisions of BCP 78 and BCP 79.
        </t>""",
        """<t>
        Internet-Drafts are working documents of the Internet Engineering Task
        Force (IETF). Note that other groups may also distribute working
        documents as Internet-Drafts. The list of current Internet-Drafts is
        at <eref target="{scheme}://datatracker.ietf.org/drafts/current/"/>.
        </t>""",
        """<t>
        Internet-Drafts are draft documents valid for a maximum of six months
        and may be updated, replaced, or obsoleted by other documents at any
        time. It is inappropriate to use Internet-Drafts as reference
        material or to cite them other than as "work in progress."
        </t>""",
        """<t>
        This Internet-Draft will expire on {expiration_date}.
        </t>""",
    ]



# --------------------------------------------------
# Copyright The IETF Trust 2011, All Rights Reserved
# --------------------------------------------------

# Python libs
import calendar

try:
    import debug
    assert debug
except ImportError:
    pass

# Local libs
from xml2rfc.writers.base import BaseRfcWriter, default_options
from xml2rfc.writers.raw_txt import RawTextRfcWriter
import xml2rfc.utils
import xml2rfc.log

class PaginatedTextRfcWriter(RawTextRfcWriter):
    """ Writes to a text file, paginated with headers and footers

        The page width is controlled by the *width* parameter.
    """

    def __init__(self, xmlrfc, width=72, quiet=None, options=default_options,
                               date=None, omit_headers=False):
        if not quiet is None:
            options.quiet = quiet
        RawTextRfcWriter.__init__(self, xmlrfc, options=options, date=date)
        self.left_header = ''
        self.center_header = ''
        self.right_header = ''
        self.left_footer = ''
        self.center_footer = ''
        self.break_hints = {}
        self.heading_marks = {}
        self.paged_toc_marker = 0
        self.page_line = 0
        self.omit_headers = omit_headers
        self.page_end_blank_lines = 2
        # Don't permit less than this many lines of a broken paragraph at the
        # top of a page:
        self.widow_limit = self.get_numeric_pi('widowlimit', default=2)
        # Don't permit less than this many lines of a broken paragraph at the
        # bottom of a page:
        self.orphan_limit = self.get_numeric_pi('orphanlimit', default=2)

    def _make_footer_and_header(self, page, final=False):
        tmp = []
        tmp.append(xml2rfc.utils.justify_inline(self.left_footer,
                                                self.center_footer,
                                                '[Page ' + str(page) + ']'))
        if not final:
            tmp.append('\f')
            tmp.append(xml2rfc.utils.justify_inline(self.left_header,
                                                    self.center_header,
                                                    self.right_header))
        return tmp

    def _vspace(self, num=0):
        """ <vspace> needs to allow for forcing page breaks """
        if num > 51:
            self._set_break_hint(-1, 'break')
            num = 1
        self._lb(num=num)
            
    def _set_break_hint(self, needLines, type, where=-1):
        """ Use this function to set break hints since it will do all of the
            necessary checks to see that we don't override a stronger hint
        """
        if where == -1:
            where = len(self.buf)
        if where in self.break_hints:
            need, ptype = self.break_hints[where]
            if ptype == 'break':
                # breaks always win
                return
            if ptype == 'raw':
                # raw is better than text
                type = ptype
            # Extend the number of lines if greater
            if need > needLines:
                needLines = need
        self.break_hints[where] = (needLines, type)
            

    def write_with_break_hint(self, writer, type, *args, **kwargs):
        """A helper function to wrap a fragment writer in code to a page break
        hint.  This function also takes care to preserve a stronger break type
        so that it's not overwritten with a weaker one."""
        begin = len(self.buf)
        writer(self, *args, **kwargs)
        needLines = len(self.buf) - begin
        for line in self.buf[begin:]:
            if self.IsFormatting(line):
                needLines -= 1
        self._set_break_hint(needLines, type, begin)

    def needLines(self, count):
        """Deal with the PI directive needLines"""
        if isinstance(count, str):
            if count.isdigit():
                count = int(count)
            else:
                xml2rfc.log.warn('Expected a numeric value for needLines, but got "%s".  Using 3.' % count)
                count = 3
        try:
            if count < 0:
                self._set_break_hint(1, 'break', len(self.buf))
            else:
                self._set_break_hint(count, 'raw', len(self.buf))
        except ValueError as e:
            if not self.indexmode:
                xml2rfc.log.warn('%s, in processing instruction needlines="%s"' % (str(e).capitalize(), count))

    # Here we override some methods to mark line numbers for large sections.
    # We'll store each marking as a dictionary of line_num: section_length.
    # This way we can step through these markings during writing to
    # preemptively construct appropriate page breaks.
    # Only those things which are supposed to be widowed or orphaned use 'txt'
    def write_figure(self, *args, **kwargs):
        """ Override base writer to add a marking """
        self.write_with_break_hint(BaseRfcWriter.write_figure, 'raw', *args, **kwargs)

    def write_table(self, *args, **kwargs):
        """ Override base writer to add a marking """
        self.write_with_break_hint(BaseRfcWriter.write_table, 'raw', *args, **kwargs)

    def write_raw(self, *args, **kwargs):
        """ Override text writer to add a marking """
        self.write_with_break_hint(RawTextRfcWriter.write_raw, 'raw', *args, **kwargs)

    def write_text(self, *args, **kwargs):
        """ Override text writer to add a marking """
        self.write_with_break_hint(RawTextRfcWriter.write_text, 'txt', *args, **kwargs)

    def write_ref_element(self, *args, **kwargs):
        """ Override text writer to add a marking """
        self.write_with_break_hint(RawTextRfcWriter.write_ref_element, 'raw', *args, **kwargs)
        
    def _force_break(self):
        """ Force a pagebreak at the current buffer position. Not used yet, waiting
        for markup that indicates a forced page break to be defined"""
        self._set_break_hint(-1, 'break', len(self.buf))
        
    def _toc_size_hint(self):
        return len(self.write_toc(paging=True))
    
    def _iref_size_hint(self):
        return len(self.write_iref_index())

    
    # ------------------------------------------------------------------------
    
    def write_heading(self, text, bullet='', autoAnchor=None, anchor=None, \
                      level=1):
        # Store the line number of this heading with its unique anchor, 
        # to later create paging info
        begin = len(self.buf)
        self.heading_marks[begin] = autoAnchor
        RawTextRfcWriter.write_heading(self, text, bullet=bullet, \
                                       autoAnchor=autoAnchor, anchor=anchor, \
                                       level=level)
        # Reserve room for a blankline and some lines of section content
        # text, in order to prevent orphan headings
        #
        # We've now written a blank line before the heading, followed by
        # the heading.  Next will be a blank line, followed by the section
        # content.  We thus need (begin - end) + 1 + orphan_limit lines, but
        # if we ask for only that, we'll end up moving the text, to the next
        # page, without moving the section heading.  So we ask for one more
        self.orphan_limit = self.get_numeric_pi('orphanlimit', default=self.orphan_limit)
        orphanlines = self.get_numeric_pi('sectionorphan', default=self.orphan_limit+2)
        end = len(self.buf)
        needed = end - begin + orphanlines
        self._set_break_hint(needed, 'raw', begin)


    def pre_rendering(self):
        """ Prepares the header and footer information """
        # Raw textwriters preprocessing will replace unicode with safe ascii
        RawTextRfcWriter.pre_rendering(self)

        # Discard hints and marks from indexing pass
        self.break_hints = {}
        self.heading_marks = {}

        if self.pis['private']:
            self.left_header = ''
        elif self.draft:
            self.left_header = 'Internet-Draft'
        else:
            self.left_header = 'RFC %s' % self.r.attrib.get('number', '')
        title = self.r.find('front/title')
        if title is not None:
            self.center_header = title.attrib.get('abbrev', title.text)
        date = self.r.find('front/date')
        if date is not None:
            month = date.attrib.get('month', '')
            if month.isdigit():
                month = calendar.month_name[int(month)]
            year = date.attrib.get('year', '')
            self.right_header = month + ' ' + year
        authors = self.r.findall('front/author')
        surnames = list(filter(None, [ a.get('asciiSurname', a.get('surname')) for a in authors ]))
        if len(surnames) == 1:
            self.left_footer = surnames[0]
        elif len(surnames) == 2:
            self.left_footer = '%s & %s' % (surnames[0], surnames[1],)
        elif len(surnames) > 2:
            self.left_footer = '%s, et al.' % surnames[0]
        if self.draft:
            self.center_footer = 'Expires %s' % self.expire_string
        else:
            self.center_footer = self.boilerplate.get(
                                 self.r.attrib.get('category', ''), '(Category')

        # Check for PI override
        self.center_footer = self.pis.get('footer', self.center_footer)
        self.left_header = self.pis.get('header', self.left_header)

    def page_break(self, final=False):
        # remove header if nothing on the page
        if final and self.page_length == 1:
            while len(self.output) > 0:
                p = self.output.pop()
                if p == '\f':
                    break;
            return
        if self.omit_headers:
            self.output.append('\f')
        else:
            self.output.append('')
            self.output.append('')
            self.output.append('')
            self.output.extend(self._make_footer_and_header(self.page_num, final))
        if not final:
            self.output.append('')
            self.output.append('')

        self.page_length = 1
        self.page_num += 1

    def emit(self, text):
        """Write text to the output buffer if it's not just a blank
           line at the top of the page"""
        if isinstance(text, type('')) or isinstance(text, type(u'')):
            if self.page_length == 1 and text.strip() in ['', u'']:
                return 
            self.output.append(text)
            self.page_length += 1
        elif isinstance(text, list):
            if self.page_length == 1:
                for line in text:
                    self.emit(line)
                return
            self.output.extend(text)
            self.page_length += len(text)
        else:
            raise TypeError("a string or a list of strings is required")

    def IsFormatting(self, line):
        return False

    def post_rendering(self):
        """ Add paging information to a secondary buffer """
        # Counters    
        self.page_length = 0
        self.page_num = 1
        max_page_length = 51
        lines_until_break = -1


        # Maintain a list of (start, end) pointers for elements to re-insert
        toc_pointers = []
        toc_prev_start = 0     
        iref_pointers = []
        iref_prev_start = 0

        for line_num, line in enumerate(self.buf):
            if line_num == self.toc_marker and self.toc_marker > 0:
                # Don't start ToC too close to the end of the page
                if self.pis['tocpagebreak'] == 'yes' or self.page_length + 10 >= max_page_length:
                    remainder = max_page_length - self.page_length - self.page_end_blank_lines
                    self.emit([''] * remainder)
                    self.page_break()

                # Insert a dummy table of contents here
                toc_prev_start = len(self.output)
                preliminary_toc = self.write_toc(paging=True)
                for l in preliminary_toc:
                    if self.page_length + self.page_end_blank_lines >= max_page_length:
                        # Store a pair of TOC pointers
                        toc_pointers.append((toc_prev_start, len(self.output)))
                        # New page
                        self.page_break()
                        toc_prev_start = len(self.output)
                    # Write dummy line
                    self.emit(l)
                # Store last pair of toc pointers
                toc_pointers.append((toc_prev_start, len(self.output)))
                
            if line_num == self.iref_marker and self.iref_marker > 0:
                # Don't start Index too close to the end of the page
                if self.page_length + 10 >= max_page_length:
                    remainder = max_page_length - self.page_length - self.page_end_blank_lines
                    self.emit([''] * remainder)
                    self.page_break()

                # Add page number for index
                item = self._getItemByAnchor('rfc.index')
                if item:
                    item.page = self.page_num
                # Insert a dummy iref here
                iref_prev_start = len(self.output)
                preliminary_iref_index = self.write_iref_index()
                for l in preliminary_iref_index:
                    if self.page_length + self.page_end_blank_lines >= max_page_length:
                        # Store a pair of pointers
                        iref_pointers.append((iref_prev_start, len(self.output)))
                        # New page
                        self.page_break()
                        iref_prev_start = len(self.output)
                    # Write dummy line
                    self.emit(l)
                # Store last pair of pointers
                iref_pointers.append((iref_prev_start, len(self.output)))

            if line_num in self.break_hints:
                # If this size hint exceeds the rest of the page, or is set
                # to -1 (a forced break), insert a break.

                available = max_page_length - (self.page_length + self.page_end_blank_lines)
                needed, text_type = self.break_hints[line_num]

                blanks = 0
                i = 0
                while ((line_num+i < len(self.buf)) and
		       (self.buf[line_num+i].strip() == "" or
                        self.IsFormatting(self.buf[line_num+i]))):
                    if not self.IsFormatting(self.buf[line_num+i]):
                        blanks += 1
                    i += 1
                if blanks > 0:
                    # discount initial blank line in what we're about to
                    # write when considering whether we're about to create
                    # orphans or widows
                    available -= blanks
                    needed -= blanks

                if (text_type == "break"
                    or (text_type == "raw" and needed > available and needed < max_page_length-self.page_end_blank_lines )):
                    lines_until_break = 0

                elif (self.pis['autobreaks'] == 'yes'
                      and needed > available
                      and needed < max_page_length-self.page_end_blank_lines
                      and (needed-available < self.widow_limit or available < self.orphan_limit) ):
                    lines_until_break = available
                    if available == needed - 1:
                        lines_until_break -= 1
                    if lines_until_break < self.page_end_blank_lines:
                        lines_until_break = 0
                    if lines_until_break > 0 and blanks > 0:
                        # put back that blank line since we are going to emit it
                        lines_until_break += blanks

            if lines_until_break == 0:
                # Insert break
                remainder = max_page_length - self.page_length - self.page_end_blank_lines
                self.emit([''] * remainder)
            if not self.IsFormatting(line):
                lines_until_break -= 1

            if self.page_length + self.page_end_blank_lines >= max_page_length:
                # New page
                self.page_break()

            self.emit(line)

            # Store page numbers for any marked elements
            if line_num in self.heading_marks:
                item = self._getItemByAnchor(self.heading_marks[line_num])
                if item:
                    item.page = self.page_num
            if line_num in self.iref_marks:
                for item, subitem in self.iref_marks[line_num]:
                    # Store pages in item unless there are subitems
                    if subitem:
                        self._iref_index[item].subitems[subitem].pages.append(self.page_num)
                    else:
                        self._iref_index[item].pages.append(self.page_num)

        # Write final footer
        if self.page_length > 1:
            remainder = max_page_length - self.page_length - self.page_end_blank_lines
            self.emit([''] * remainder)
        self.page_break(final=True)
        
        # Now we need to go back into the buffer and insert the real table 
        # of contents and iref based on the pointers we created
        if len(toc_pointers) > 0:
            tocbuf = self.write_toc(paging=True)
            ptr, end = toc_pointers.pop(0)
            for line in tocbuf:
                if self.output[ptr] != '' and line == '':
                    continue
                self.output[ptr] = line
                ptr += 1
                if ptr >= end:
                    if len(toc_pointers) > 0:
                        ptr, end = toc_pointers.pop(0)
                    else:
                        break

        if len(iref_pointers) > 0:
            irefbuf = self.write_iref_index()
            ptr, end = iref_pointers.pop(0)
            for line in irefbuf:
                if self.output[ptr] != '' and line == '':
                    continue
                self.output[ptr] = line
                ptr += 1
                if ptr >= end:
                    if len(iref_pointers) > 0:
                        ptr, end = iref_pointers.pop(0)
                    else:
                        break

    def post_process_lines(self, lines):
        return RawTextRfcWriter.post_process_lines(self, lines)
